/*
 * @Description: 
 * @Author: luzc
 * @Date: 2021-04-12 11:01:59
 */
'use strict';

const fs = require('fs');
const path = require('path');
const webpack = require('webpack');
const HtmlWebpackPlugin = require('html-webpack-plugin');
const TerserPlugin = require('terser-webpack-plugin');
const CssMinimizerPlugin = require('css-minimizer-webpack-plugin');
const MiniCssExtractPlugin = require('mini-css-extract-plugin');
const PnpWebpackPlugin = require('pnp-webpack-plugin');
const { WebpackManifestPlugin } = require('webpack-manifest-plugin');
// const ModuleScopePlugin = require('react-dev-utils/ModuleScopePlugin');
const getCSSModuleLocalIdent = require('react-dev-utils/getCSSModuleLocalIdent');
const InterpolateHtmlPlugin = require('react-dev-utils/InterpolateHtmlPlugin');
const WatchMissingNodeModulesPlugin = require('react-dev-utils/WatchMissingNodeModulesPlugin');
const { getLess } = require("@zougt/some-loader-utils");
const postcssNormalize = require('postcss-normalize');
const reactRefreshOverlayEntry = require.resolve(
  'react-dev-utils/refreshOverlayInterop'
);
const paths = require('./paths');
const modules = require('./modules');
const getClientEnvironment = require('./env');

const appPackageJson = require(paths.appPackageJson);
const useTypeScript = fs.existsSync(paths.appTsConfig);

// 自定义配置
const {
  alias,
  addWebpackPlugins,
  addBabelPlugins,
  externals,
  pxtorem,
  modifyVars,
  isUmdLibrary,
  themesConfig
} = paths.appConfig;

module.exports = (webpackEnv) => {
  const isEnvDevelopment = webpackEnv === 'development';
  const isEnvProduction = webpackEnv === 'production';
  const env = getClientEnvironment(paths.publicUrlOrPath.slice(0, -1));

  const hasJsxRuntime = (() => {
    if (process.env.DISABLE_NEW_JSX_TRANSFORM === 'true') {
      return false;
    }
    try {
      require.resolve('react/jsx-runtime');
      return true;
    } catch (e) {
      return false;
    }
  })();

  const getStyleLoaders = (importLoaders, isModules = false, isLessLoader = false) => {
    const loaders = [
      isEnvDevelopment && ({
        loader: require.resolve('style-loader'),
        options: {
          attributes: {
            class: `app-${appPackageJson.name}-styles`,
          }
        }
      }),
      isEnvProduction && (
        {
          loader: MiniCssExtractPlugin.loader
        }
      ),
      {
        loader: require.resolve('css-loader'),
        options: {
          importLoaders,
          sourceMap: isEnvDevelopment,
          ...(
            isModules
              ? {
                modules: {
                  getLocalIdent: getCSSModuleLocalIdent
                }
              }
              : {}
          )
        }
      },
      {
        loader: require.resolve('postcss-loader'),
        options: {
          sourceMap: isEnvDevelopment,
          postcssOptions: {
            ident: 'postcss',
            plugins: (() => [
              require('postcss-import'),
              require('postcss-flexbugs-fixes'),
              require('postcss-preset-env')({
                autoprefixer: {
                  flexbox: 'no-2009'
                },
                stage: 3
              }),
              postcssNormalize(),
              pxtorem && require('postcss-pxtorem')({
                rootValue: 16,
                propList: ['*'],
                ...(
                  Object.prototype.toString.call(pxtorem) === '[object Object]'
                    ? pxtorem
                    : {}
                )
              })
            ].filter(Boolean))()
          }
        }
      }
    ].filter(Boolean);

    if (isLessLoader) {
      loaders.push(
        {
          loader: require.resolve('resolve-url-loader'),
          options: {
            sourceMap: isEnvDevelopment,
            root: paths.appSrc,
            removeCR: true,
          }
        },
        {
          loader: require.resolve('less-loader'),
          options: {
            sourceMap: isEnvDevelopment,
            lessOptions: {
              javascriptEnabled: true,
              ...(
                Object.prototype.toString.call(modifyVars) === '[object Object]'
                  ? { modifyVars }
                  : {}
              )
            },
            // 多主题
            ...(Array.isArray(themesConfig) ? {
              implementation: getLess({
                getMultipleScopeVars: (lessOptions) => themesConfig,
              })
            } : {})
          }
        }
      )
    }
    return loaders;
  }

  return {
    mode: isEnvProduction ? 'production' : isEnvDevelopment && 'development',
    bail: isEnvProduction,
    devtool: isEnvProduction ? false : 'cheap-module-source-map',
    entry: [paths.appIndexJs],
    output: {
      path: paths.appBuild,
      pathinfo: isEnvDevelopment,
      filename: isEnvProduction
        ? 'static/js/[name].[contenthash:8].js'
        : 'static/js/[name].js',
      chunkFilename: isEnvProduction
        ? 'static/js/[name].[contenthash:8].chunk.js'
        : 'static/js/[name].chunk.js',
      publicPath: paths.publicUrlOrPath,
      globalObject: 'this',
      devtoolModuleFilenameTemplate: isEnvProduction
        ? info =>
          path
            .relative(paths.appSrc, info.absoluteResourcePath)
            .replace(/\\/g, '/')
        : (info => path.resolve(info.absoluteResourcePath).replace(/\\/g, '/')),
      // 开启 umd 模式
      ...(
        isUmdLibrary
          ? {
            library: `micro-app-${appPackageJson.name}`,
            libraryTarget: 'umd'
          }
          : {}
      )
    },
    externals,
    cache: {
      type: 'filesystem',
      cacheDirectory: paths.cacheDir
    },
    optimization: {
      minimize: isEnvProduction,
      minimizer: [
        new TerserPlugin({
          parallel: true,
          terserOptions: {
            ecma: 8,
            compress: {
              warnings: false,
              drop_debugger: true,
              comparisons: false,
              inline: 2,
              pure_funcs: ["console.log"], // 删除所有的console.log
            },
            mangle: {
              safari10: true,
            },
            output: {
              ecma: 5,
              comments: false,
              ascii_only: true,
            },
          }
        }),
        new CssMinimizerPlugin(),
      ],
      splitChunks: {
        chunks: 'all',
        name: false
      }
    },
    resolve: {
      extensions: paths.moduleFileExtensions
        .map(ext => `.${ext}`)
        .filter(ext => useTypeScript || !ext.includes('ts')),
      modules: ['node_modules', paths.appNodeModules].concat(
        modules.additionalModulePaths || []
      ),
      alias: {
        ...(modules.webpackAliases || {}),
        ...(alias || {})
      },
      plugins: [
        PnpWebpackPlugin
      ]
    },
    resolveLoader: {
      plugins: [
        PnpWebpackPlugin.moduleLoader(module),
      ],
    },
    module: {
      parser: {
        javascript: {
          requireEnsure: true,
        }
      },
      rules: [
        {
          oneOf: [
            {
              test: [/\.avif$/],
              loader: require.resolve('url-loader'),
              options: {
                limit: 10000,
                mimetype: 'image/avif',
                name: 'static/media/[name].[hash:8].[ext]',
              },
            },
            {
              test: [/\.bmp$/, /\.gif$/, /\.jpe?g$/, /\.png$/],
              loader: require.resolve('url-loader'),
              options: {
                limit: 10000,
                name: 'static/media/[name].[hash:8].[ext]',
              },
            },
            {
              test: /\.(js|mjs|jsx|ts|tsx)$/,
              include: paths.appSrc,
              loader: require.resolve('babel-loader'),
              options: {
                customize: require.resolve(
                  'babel-preset-react-app/webpack-overrides'
                ),
                presets: [
                  [
                    require.resolve('babel-preset-react-app'),
                    {
                      runtime: hasJsxRuntime ? 'automatic' : 'classic',
                    },
                  ],
                ],
                plugins: [
                  [
                    require.resolve('babel-plugin-named-asset-import'),
                    {
                      loaderMap: {
                        svg: {
                          ReactComponent:
                            '@svgr/webpack?-svgo,+titleProp,+ref![path]',
                        },
                      },
                    },
                  ],
                  ...(addBabelPlugins || [])
                  // isEnvDevelopment && require.resolve('react-refresh/babel'),
                ].filter(Boolean),
                cacheDirectory: true,
                cacheCompression: false,
                compact: isEnvProduction,
              },
            },
            {
              test: /\.(js|mjs)$/,
              exclude: /@babel(?:\/|\\{1,2})runtime/,
              loader: require.resolve('babel-loader'),
              options: {
                babelrc: false,
                configFile: false,
                compact: false,
                presets: [
                  [
                    require.resolve('babel-preset-react-app/dependencies'),
                    { helpers: true },
                  ],
                ],
                cacheDirectory: true,
                cacheCompression: false,
                sourceMaps: isEnvDevelopment,
                inputSourceMap: isEnvDevelopment,
              },
            },
            {
              test: /\.css$/,
              exclude: /\.module\.css$/,
              sideEffects: true,
              use: getStyleLoaders(1, false, false)
            },
            {
              test: /\.module\.css$/,
              use: getStyleLoaders(1, true, false)
            },
            {
              test: /\.less$/,
              exclude: /\.module\.less$/,
              sideEffects: true,
              use: getStyleLoaders(3, false, true)
            },
            {
              test: /\.module\.less$/,
              use: getStyleLoaders(3, true, true)
            },
            {
              exclude: [/\.(js|mjs|jsx|ts|tsx)$/, /\.html$/, /\.json$/],
              use: (info) => (
                info.compiler !== 'HtmlWebpackCompiler'
                  ? [
                    {
                      loader: require.resolve('file-loader'),
                      options: {
                        name: 'static/media/[name].[hash:8].[ext]',
                        esModule: false,
                      },
                    }
                  ]
                  : []
              )
            },
          ]
        }
      ]
    },
    plugins: [
      new webpack.ProgressPlugin(),

      new webpack.ProvidePlugin({
        process: 'process/browser',
      }),

      new HtmlWebpackPlugin(
        Object.assign(
          {},
          {
            inject: true,
            template: paths.appHtml,
          },
          isEnvProduction
            ? {
              minify: {
                removeComments: true,
                collapseWhitespace: true,
                removeRedundantAttributes: true,
                useShortDoctype: true,
                removeEmptyAttributes: true,
                removeStyleLinkTypeAttributes: true,
                keepClosingSlash: true,
                minifyJS: true,
                minifyCSS: true,
                minifyURLs: true,
              },
            }
            : undefined
        )
      ),

      new InterpolateHtmlPlugin(HtmlWebpackPlugin, env.raw),

      new webpack.DefinePlugin(env.stringified),

      isEnvProduction && new MiniCssExtractPlugin({
        filename: 'static/css/[name].[contenthash:8].css',
        chunkFilename: 'static/css/[name].[contenthash:8].chunk.css',
        ignoreOrder: true
      }),

      isEnvDevelopment && new webpack.HotModuleReplacementPlugin(),

      isEnvDevelopment && new WatchMissingNodeModulesPlugin(paths.appNodeModules),

      new webpack.IgnorePlugin({
        resourceRegExp: /^\.\/locale$/,
        contextRegExp: /moment$/,
      }),

      new WebpackManifestPlugin({
        fileName: 'asset-manifest.json',
        publicPath: paths.publicUrlOrPath,
        generate(seed, files, entrypoints) {
          const manifestFiles = files.reduce((manifest, file) => {
            manifest[file.name] = file.path;
            return manifest;
          }, seed);
          const entrypointFiles = entrypoints.main.filter(
            fileName => !fileName.endsWith('.map')
          );
          return {
            files: manifestFiles,
            entrypoints: entrypointFiles,
          };
        }
      }),

      ...(addWebpackPlugins || [])
    ].filter(Boolean)
  }
}